package net.larsan.test.jetty;

import static java.util.concurrent.TimeUnit.SECONDS;
import static org.eclipse.jetty.servlet.ServletContextHandler.SESSIONS;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.CountDownLatch;

import org.apache.log4j.Logger;
import org.eclipse.jetty.server.Connector;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.nio.SelectChannelConnector;
import org.eclipse.jetty.servlet.ServletContextHandler;
import org.eclipse.jetty.servlet.ServletHolder;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import de.roderick.weberknecht.WebSocketConnection;
import de.roderick.weberknecht.WebSocketEventHandler;
import de.roderick.weberknecht.WebSocketException;
import de.roderick.weberknecht.WebSocketMessage;

/*
 * This test opens a Jetty server and mounts the echo servlet
 * under "/echo". It then tries to send consecutive messages to 
 * the servlet and checks that they get back ok. 
 * 
 * Current behaviour: 1 and 3 messages works, but 4 messages fails. 
 * 
 * Expected behaviour: Any number of consecutive messages should work.
 * 
 * This mimicks a real behaviour with a large server where consecutive
 * messages blocks the server. 
 * 
 * Workarounds: Use old SocketConnector or put a Thread.sleep(100) between 
 * the send messages.

 */
public class TestJettyWebSockets {

	private static final long MAX_WAIT = 5;

	private final Logger log = Logger.getLogger(getClass());
	
	private Server server;

	@Before
	public void init() throws Exception {
		server = new Server(); 
		// Switch to SocketConnectory to make it work again
		Connector connector = new SelectChannelConnector();
		connector.setPort(8080);
		server.setConnectors(new Connector[] { connector });
		// Setup servlet
		ServletContextHandler context = new ServletContextHandler(SESSIONS);
		context.addServlet(new ServletHolder(new EchoSocketServlet()), "/echo");
        context.setContextPath("/");
        server.setHandler(context);
        // Start
        log.info("Jetty starting");
        server.start();
	}
	
	@After
	public void destroy() throws Exception {
		server.stop();
	}
	
	@Test
	public void testSingleShot() throws Exception {
		testMultiple(1); // Works...	
	}
	
	@Test
	public void testThreeConsecutive() throws Exception {
		testMultiple(3); // Works...
	}
	
	@Test
	public void testFourConsecutive() throws Exception {
		testMultiple(4); // Fails with SelectChannelConnector, works with SocketConnector
	}
	
	
	// --- PRIVATE METHODS --- //

	private void testMultiple(int count) throws WebSocketException, URISyntaxException, InterruptedException {
		Handler h = new Handler(count);
		WebSocketConnection websocket = createWebSocket(h);
		List<String> list = new LinkedList<String>();
		// Send messages and store for later checks
		for (int i = 0; i < count; i++) {
			String msg = String.valueOf(i);
			websocket.send(msg);
			list.add(msg);
		}
		// Wait for the latch, then check all messages
		if(h.latch.await(MAX_WAIT, SECONDS)) {
			for (int i = 0; i < count; i++) {
				assertEquals(list.get(i), h.incoming.get(i));
			}
		} else {
			fail("Message dropped!");
		}
	}

	private WebSocketConnection createWebSocket(Handler h) throws WebSocketException, URISyntaxException {
		WebSocketConnection websocket = new WebSocketConnection(new URI("ws://localhost:8080/echo"));
		websocket.setEventHandler(h);
		websocket.connect();
		return websocket;
	}
	

	// --- PRIVATE CLASSES --- //
	
	private class Handler implements WebSocketEventHandler {
		
		private List<String> incoming = new LinkedList<String>();
		private CountDownLatch latch;
		
		public Handler(int expected) {
			/*
			 *  The count-down latch is used to block until all
			 *  expected messages have arrived. 
			 */
			latch = new CountDownLatch(expected);
		}

		public void onClose() {
			log.info("Web socket closed");
		}
		
		public void onMessage(WebSocketMessage msg) {
			// Store and count down
			incoming.add(msg.getText());
			latch.countDown();
		}
		
		public void onOpen() {
			log.info("Web socket opened");
		}
	}
}
